use Test::More;
use Mxpress::PDF;
use utf8;

my $pdf = Mxpress::PDF->new_pdf('test', 
	page => { background => '#000' },
	title => { font => { colour => '#f00' } },
	subtitle => { padding => 5, font => { colour => '#0ff' } },
	text => { padding => 5, font => { colour => '#fff' } }
)->add_page->title->add('This is a title', position => [5, 297])->subtitle->add('This is a subtitle.')->text->add('This is some text.')->save();

ok(1);

done_testing();
